import os
import json
import sys
import platform
import shutil
from typing import Dict, Union

import dew.args
from dew.args import Arguments, CommandType
from dew.buildoptions import BuildOptionsCache, BuildOptions
from dew.depstate import DependencyStateController
from dew.dewfile import DewFileParser
from dew.exceptions import DewError
from dew.projectprocessor import ProjectProcessor
from dew.storage import StorageController
from dew.view import View


def main() -> int:
    parser = dew.args.make_argparser()
    args = Arguments()

    # noinspection PyTypeChecker
    parser.parse_args(namespace=args)

    view = View()
    view.show_verbose = True if args.verbose else False

    if args.version:
        print(f'dew {dew.VERSION_MAJOR}.{dew.VERSION_MINOR}.{dew.VERSION_PATCH}')
        return 0

    if args.command == CommandType.UPDATE:
        return update(args, view)
    if args.command == CommandType.BOOTSTRAP:
        return boostrap()
    if args.command == CommandType.CLEAN:
        return clean(args, view)

    view.error(f'Don\'t know how to handle command {args.command.value}! This is a bug.')
    return 1


def get_storage(args: Arguments) -> StorageController:
    if args.output_path:
        storage_path = args.output_path
    else:
        storage_path = os.path.join(os.getcwd(), '.dew')

    return StorageController(storage_path)


def update(args: Arguments, view: View) -> int:
    storage = get_storage(args)
    storage.ensure_directories_exist()

    options_cache = BuildOptionsCache(storage)
    options = options_cache.get_options()
    if not options:
        options = BuildOptions()

    apply_argument_options(options, args)

    if args.defines is not None:
        for define in args.defines:
            options.options[define] = True

    options.cmake_executable = args.cmake_executable

    options.options.update(get_builtin_options())

    options_cache.save_options(options)

    invalid_options = options.get_invalid_options()
    if len(invalid_options) > 0:
        view.error('The following options are either unset or set to invalid values:\n{0}\n'
                   'Please specify the correct values as arguments.'.format('\n'.join(invalid_options)))
        return 64

    dewfile_path = 'dewfile.json'
    if args.dewfile:
        dewfile_path = args.dewfile

    dewfile = dew.dewfile.parse_dewfile_with_local_overlay(dewfile_path)

    depstates = DependencyStateController(storage)
    depstates.load()

    project_processor = ProjectProcessor(storage, options, view, depstates)
    project_processor.set_data(dewfile)

    success = True

    try:
        project_processor.process()
    except DewError:
        view.error('Dew did not finish successfully :(')
        success = False

    if success and args.update_dummy_file:
        with open(args.update_dummy_file, 'a'):
            os.utime(args.update_dummy_file, None)

    depstates.save()

    return 0 if success else 1


def clean(args: Arguments, view: View) -> int:
    view.error('Not implemented yet!')
    return 1


def boostrap() -> int:
    src_path = os.path.join(__file__, '..', '..', 'cmake', 'dew.cmake')
    src_path = os.path.normpath(src_path)
    dest_path = os.path.join(os.getcwd(), 'dew.cmake')
    shutil.copy(src_path, dest_path)
    return 0


def apply_argument_options(options: BuildOptions, args) -> None:
    if args.cmake_generator:
        options.cmake_generator = args.cmake_generator


def get_builtin_options() -> Dict[str, Union[str, bool]]:
    options = {}
    if platform.system() == 'Darwin':
        if len(platform.mac_ver()[0]) > 0:
            options['DEW_PLATFORM_MACOS'] = True
    return options


sys.exit(main())



import sys

from typing import Optional, Tuple

import git.repo


def update_repo(url: str, destination_dir: str, ref: Optional[str]=None) -> None:
    repo, origin = setup_repo(url, destination_dir)
    fetch_and_checkout(repo, origin, ref)


def fetch_and_checkout(repo: git.Repo, origin: git.Remote, ref: Optional[str]=None) -> None:
    if not ref:
        ref = 'master'

    fetch_info = None
    need_to_fetch_all = False
    try:
        fetch_info = origin.fetch(refspec=ref)[0]
    except git.GitCommandError as e:
        if 'unadvertised object' in e.stderr:
            # The ref might have been an unadvertised object (e.g. github does not advertise individual commits and does
            # now allow fetching an individual commit). If this is the case, we must fetch everything.
            need_to_fetch_all = True
        else:
            raise

    if need_to_fetch_all:
        origin.fetch()
        head_commit = ref
    else:
        head_commit = fetch_info.ref.commit

    head = repo.create_head(path='dew-head', commit=head_commit)
    head.checkout(force=True)

    failed_fetching_submodules = False

    for submodule in repo.submodules:
        submodule.update(init=True)


def setup_repo(url: str, destination_dir: str) -> Tuple[git.Repo, git.Remote]:
    repo: git.Repo = None
    origin: git.Remote = None
    try:
        repo = git.repo.Repo(destination_dir)
    except (git.NoSuchPathError, git.InvalidGitRepositoryError):
        pass

    if repo is None:
        repo = git.repo.Repo.init(path=destination_dir, mkdir=True)

    for remote in repo.remotes:
        if remote.name == 'origin':
            origin = remote

    if origin is None:
        origin = repo.create_remote('origin', url)
    else:
        invalid_urls = set()
        for current_url in origin.urls:
            if current_url != url:
                invalid_urls.add(current_url)
        for invalid_url in invalid_urls:
            origin.set_url(url, invalid_url)

    return repo, origin



import os
import shutil
from typing import Set, Iterable, Dict

from dew.buildoptions import BuildOptions
from dew.dependencygraph import DependencyGraph
from dew.dependencyprocessor import DependencyProcessor
from dew.depstate import DependencyStateController
from dew.dewfile import DewFile
from dew.exceptions import BuildError
from dew.storage import StorageController
from dew.view import View


class ProjectProcessor(object):

    def __init__(self, storage: StorageController, options: BuildOptions, view: View,
                 depstates: DependencyStateController):
        self.storage = storage
        self.root_dewfile = None
        self.options = options
        self.view = view
        self.depstates = depstates

    def set_data(self, dewfile: DewFile):
        self.root_dewfile = dewfile

    def process(self):
        dewfile_stack = [(self.root_dewfile, None)]

        # Dependency processors by label
        dependency_processors = {}

        graph = DependencyGraph()
        deps_needing_rebuild: Set[str] = set()

        while len(dewfile_stack) > 0:
            dewfile, parent_name = dewfile_stack.pop()
            for dep in dewfile.dependencies:
                dep_processor = DependencyProcessor(self.storage, self.view, dep, dewfile, self.options)
                label = self.get_label(dep.name, dep_processor.get_version())
                dep_processor.set_label(label)
                dependency_processors[label] = dep_processor

                graph.add_dependency(label, parent_name)

                if self.depstates.get_state(label):
                    self.view.verbose(f'{label} is up to date.')
                    continue

                deps_needing_rebuild.add(label)

                self.view.info('Pulling dependency {0}...'.format(label))
                dep_processor.pull()

                if dep_processor.has_dewfile():
                    dewfile_stack.append((dep_processor.get_dewfile(), label))

        labels_in_order = graph.resolve()

        for label in labels_in_order:
            dep_processor = dependency_processors[label]

            if label not in deps_needing_rebuild:
                continue

            self.view.info('Building dependency {0}...'.format(dep_processor.dependency.name))

            # Prepare output prefix
            child_labels = [n.name for n in graph.nodes[label].children]
            output_prefix = self.get_isolated_prefix(label)
            shutil.rmtree(output_prefix)

            input_prefixes = [self.get_isolated_prefix(l) for l in child_labels]

            # Build and install
            dep_processor.build(output_prefix, input_prefixes)

            self.depstates.add(label)

        if len(deps_needing_rebuild) > 0:
            self.update_final_prefix(labels_in_order)

    def get_isolated_prefix(self, label: str) -> str:
        path = os.path.join(self.storage.get_output_prefix_dir(), label)
        os.makedirs(path, exist_ok=True)
        return path

    def copy_prefix(self, src_prefix: str, dst_prefix: str) -> None:
        for dirpath, dirnames, filenames in os.walk(src_prefix):
            relpath = os.path.relpath(dirpath, src_prefix)
            dst_dir = os.path.join(dst_prefix, relpath)

            for dirname in dirnames:
                os.makedirs(os.path.join(dst_dir, dirname), exist_ok=True)

            for filename in filenames:
                src_path = os.path.join(dirpath, filename)
                dst_path = os.path.join(dst_dir, filename)

                shutil.copy2(src_path, dst_path)

    def update_final_prefix(self, labels: Iterable[str]):
        success = True
        dst_prefix = self.storage.get_install_dir()
        installed_files: Dict[str, str] = {}

        for label in labels:
            src_prefix = self.get_isolated_prefix(label)

            for dirpath, dirnames, filenames in os.walk(src_prefix):
                relpath = os.path.relpath(dirpath, src_prefix)
                dst_dir = os.path.join(dst_prefix, relpath)

                for dirname in dirnames:
                    os.makedirs(os.path.join(dst_dir, dirname), exist_ok=True)

                for filename in filenames:
                    src_path = os.path.join(dirpath, filename)
                    dst_path = os.path.join(dst_dir, filename)

                    prefix_neutral_path = os.path.join(relpath, filename)

                    shutil.copy2(src_path, dst_path)

                    if prefix_neutral_path in installed_files:
                        self.view.error(f'Conflicting prefix files found! file: {prefix_neutral_path}, first occurance: {installed_files[prefix_neutral_path]}, current occurance: {label}')
                        success = False
                    else:
                        # Only copy files if we are currently succeding.
                        if success:
                            shutil.copy2(src_path, dst_path)
                        installed_files[prefix_neutral_path] = label

        if not success:
            raise BuildError(f'Failed while installing files')

    def get_label(self, name: str, version: str) -> str:
        return f'{name}_{version}'



import argparse
import enum
from typing import List


class CommandType(enum.Enum):
    UPDATE = 'update'
    BOOTSTRAP = 'bootstrap'
    CLEAN = 'clean'


class Arguments(object):
    def __init__(self) -> None:
        self.command = CommandType.UPDATE
        self.cmake_generator = ''
        self.verbose = False
        self.version = False
        self.defines: List[str] = []
        self.dewfile = ''
        self.output_path = ''
        self.update_dummy_file = ''
        self.cmake_executable = ''


def make_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('command', type=CommandType, nargs='?', default=CommandType.UPDATE)
    parser.add_argument('-v', help='Be verbose', dest='verbose', action='store_true')
    parser.add_argument('--version', action='store_true')
    parser.add_argument('-D', action='append', dest='defines', help='Define the value of an option')
    parser.add_argument('--output-path')
    parser.add_argument('--dewfile')
    parser.add_argument('--dummy-file', dest='update_dummy_file')
    parser.add_argument('--cmake-generator')
    parser.add_argument('--cmake-executable')

    return parser



import re
from typing import List, Dict, Iterable, Optional


cache_entry_pattern = re.compile(r'^\s*([^:]*)(?::(.*))?=(.*?)(\s*/{2}.*)?$')


class CMakeCacheEntry(object):
    def __init__(self) -> None:
        self.source_line = -1
        self.name = ''
        self.type = ''
        self.value = ''
        self.comment = ''


class CMakeCacheDocument(object):

    def __init__(self, lines: Iterable[str], entries: Iterable[CMakeCacheEntry]) -> None:
        self._lines: List[str] = list(lines)
        self._entries: Dict[str, CMakeCacheEntry] = {}
        for entry in entries:
            self._entries[entry.name] = entry

    def set_value(self, name: str, cmake_type: str, value: str) -> None:
        entry = self._entries.get(name)
        if entry is None:
            entry = CMakeCacheEntry()
            entry.source_line = len(self._lines)
            entry.name = name
            self._lines.append('')
            self._entries[name] = entry

        entry.type = cmake_type
        entry.value = value
        self._apply_entry(entry)

    def _apply_entry(self, entry: CMakeCacheEntry) -> None:
        text = '{0}:{1}={2}'.format(entry.name, entry.type, entry.value)
        self._lines[entry.source_line] = text

    @property
    def lines(self) -> List[str]:
        return list(self._lines)


def parse_cmake_cache(lines: Iterable[str]) -> Optional[CMakeCacheDocument]:
    entries: List[CMakeCacheEntry] = []

    for line_index, line in enumerate(lines):
        if line.strip().startswith('#'):
            continue

        match = cache_entry_pattern.match(line)
        if match is None:
            continue

        entry = CMakeCacheEntry()
        entry.source_line = line_index
        entry.name = match.group(1)
        entry.type = match.group(2)
        if entry.type is None:
            entry.type = ''
        entry.value = match.group(3)
        entry.comment = match.group(4)
        if entry.comment is None:
            entry.comment = ''

        entries.append(entry)

    doc = CMakeCacheDocument(lines, entries)
    return doc



import os

import json
from typing import List, Any, Dict


class DewFile(object):
    def __init__(self) -> None:
        self.options: List[Option] = []
        self.dependencies: List[Dependency] = []


class Option(object):
    def __init__(self) -> None:
        self.name = ''
        self.default = False
        self.cmake_option = ''


class Dependency(object):
    def __init__(self) -> None:
        self.name = ''
        self.url = ''
        self.type = ''
        self.ref = ''
        self.buildfile_dir = ''
        self.build_arguments = []
        self.fixes = []


class Fix(object):
    def __init__(self):
        self.type = ''
        self.params = {}


class DewFileParser(object):
    def __init__(self):
        self.data = None

    def set_data(self, data) -> None:
        self.data = data

    def parse(self) -> DewFile:
        options = []
        dependencies = []
        options_obj = self.data.get('options', [])
        dependencies_obj = self.data['dependencies']
        for option in options_obj:
            options.append(self.parse_option(option))
        for dep in dependencies_obj:
            dependencies.append(self.parse_dependency(dep))

        dewfile = DewFile()
        dewfile.options = options
        dewfile.dependencies = dependencies
        return dewfile

    def parse_option(self, obj: Dict[str, Any]) -> Option:
        option = Option()
        option.name = obj['name']
        option.default = bool(obj.get('default', False))
        option.cmake_option = obj.get('cmake_option', '')
        return option

    def parse_dependency(self, obj: Dict[str, Any]) -> Dependency:
        dep = Dependency()
        dep.name = obj['name']
        dep.url = obj['url']
        dep.type = obj['type']
        dep.ref = obj['ref']
        dep.buildfile_dir = obj.get('buildfile_dir', '')
        dep.execute_after_fetch = obj.get('execute_after_fetch', '')

        build_arguments = obj.get('build_arguments', [])
        if isinstance(build_arguments, str):
            build_arguments = [build_arguments]

        for arg in build_arguments:
            dep.build_arguments.append(str(arg))

        fixes_obj = obj.get('fixes', [])
        for fix in fixes_obj:
            dep.fixes.append(self.parse_fix(fix))

        return dep

    def parse_fix(self, obj) -> Fix:
        fix = Fix()
        fix.type = obj['type']
        fix.params = obj['params']
        return fix


def overlay_dewfile(base: DewFile, overlay: DewFile) -> None:
    for dependency in overlay.dependencies:
        index = -1
        for i, base_dependency in enumerate(base.dependencies):
            if base_dependency.name == dependency.name:
                index = i
                break

        if index != -1:
            base.dependencies[index] = dependency


def parse_dewfile_with_local_overlay(path: str) -> DewFile:
    dewfile_parser = DewFileParser()
    with open(path) as file:
        dewfile_data = json.load(file)
    dewfile_parser.set_data(dewfile_data)
    dewfile = dewfile_parser.parse()

    local_overlay_path = os.path.normpath(os.path.join(path, '..', 'dewfile.local.json'))
    if os.path.isfile(local_overlay_path):
        dewfile_parser = DewFileParser()
        with open(local_overlay_path) as file:
            overlaw_dewfile_data = json.load(file)
        dewfile_parser.set_data(overlaw_dewfile_data)
        local_dewfile = dewfile_parser.parse()
        overlay_dewfile(dewfile, local_dewfile)

    return dewfile



import multiprocessing
import os
from typing import Iterable

from dew.subprocesscaller import SubprocessCaller
from dew.builder import Builder
from dew.buildoptions import BuildOptions
from dew.dewfile import Dependency
from dew.view import View


class CMakeBuilder(Builder):
    def __init__(self, buildfile_dir: str, build_dir: str, install_dir: str, dependency: Dependency,
                 options: BuildOptions, caller: SubprocessCaller, view: View, prefix_paths: Iterable[str]) -> None:
        self.buildfile_dir = os.path.abspath(buildfile_dir)
        self.build_dir = os.path.abspath(build_dir)
        self.install_dir = os.path.abspath(install_dir)
        self.dependency = dependency
        self.options = options
        self.caller = caller
        self.view = view
        self.prefix_paths = list(prefix_paths)

    def get_cmake_executable(self) -> str:
        cmake_executable = self.options.cmake_executable
        if not cmake_executable:
            cmake_executable = 'cmake'
        return cmake_executable

    def build(self) -> None:
        install_dir = self.install_dir
        os.makedirs(self.build_dir, exist_ok=True)

        cmake_executable = self.get_cmake_executable()

        args = [
            cmake_executable,
            '-G', self.options.cmake_generator,
            self.buildfile_dir,
            '-DCMAKE_INSTALL_PREFIX={0}'.format(install_dir),
            '-DCMAKE_PREFIX_PATH={0}'.format(';'.join(self.prefix_paths)),
            '-DCMAKE_BUILD_TYPE=Debug',  # TODO: Support building dependencies in release mode.
            '-DDEW_CMAKE_INTEGRATION_ENABLED=OFF'
        ]
        args.extend(self.dependency.build_arguments)

        # Configure
        self.caller.call(args, cwd=self.build_dir)

        build_args = [cmake_executable, '--build', '.']
        generator = self.options.cmake_generator
        if generator in ('Unix Makefiles', 'MinGW Makefiles'):
            build_args.extend(('--', '-j', str(multiprocessing.cpu_count())))

        # Build
        self.caller.call(
            build_args,
            cwd=self.build_dir,
        )

        # Install
        self.caller.call(
            [cmake_executable, '--build', '.', '--target', 'install'],
            cwd=self.build_dir
        )



import time

import os
import shutil
import stat

from dew.dewfile import Dependency
from dew.remote import Remote


class LocalRemote(Remote):
    def __init__(self, dependency: Dependency) -> None:
        self.dependency = dependency

    def pull(self, dest_dir: str) -> None:
        if os.path.exists(dest_dir):
            if os.path.isdir(dest_dir):
                def remove_readonly(func, path, excinfo):
                    os.chmod(path, stat.S_IWRITE)
                    func(path)

                shutil.rmtree(dest_dir, onerror=remove_readonly)
            else:
                os.remove(dest_dir)
        shutil.copytree(self.dependency.url, dest_dir, copy_function=shutil.copy2)

    def get_version(self) -> str:
        # Return unix timestamp of most recent file modification time in source directory
        t = max(os.path.getmtime(root) for root, _, _ in os.walk(self.dependency.url))
        return str(int(t))



from dew import git
from dew.dewfile import Dependency
from dew.remote import Remote


class GitRemote(Remote):

    def __init__(self, dependency: Dependency) -> None:
        self.dependency = dependency

    def pull(self, dest_dir: str) -> None:
        git.update_repo(self.dependency.url, dest_dir, self.dependency.ref)

    def get_version(self) -> str:
        return self.dependency.ref



